/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TEncBinCoderCABAC.h
    \brief    binary entropy encoder of CABAC
*/

#ifndef __TCOM_QUEUE__
#define __TCOM_QUEUE__


template<class T>
class Queue
{
public:
  Queue()
  : m_aacT( new std::vector<T>[2] )
  {
    reset();
  }

  Void reset()
  {
    m_aucWrite = 0;
    m_aucRead = 0;
    m_uiReadPos = 0;
    m_uiWritePos = 0;
    m_uiPopCount = 0;
  }

  virtual ~Queue() { delete[] m_aacT; }

  class AccessError {};

  size_t push( T r )
  {
    if( m_uiWritePos < m_aacT[m_aucWrite].size() )
    {
      m_aacT[m_aucWrite][m_uiWritePos++] = r;
    }
    else if( m_aucWrite == m_aucRead )
    {
      m_aucWrite = 1 - m_aucWrite;
      m_uiWritePos = 0;
      push( r );
    }
    else
    {
      m_aacT[m_aucWrite].push_back( r );
      m_uiWritePos++;
    }
    return m_uiPopCount + size() - 1;
  }

  size_t size()
  {
    if( m_aucRead == m_aucWrite )
    {
      return m_uiWritePos - m_uiReadPos;
    }
    else
    {
      return m_uiWritePos + m_aacT[m_aucRead].size() - m_uiReadPos;
    }
  }

  Void pop()
  {
    if( size() == 0 )
    {
      throw AccessError();
    }
    if( m_uiReadPos == m_aacT[m_aucRead].size() )
    {
      m_uiReadPos = 0;
      m_aucRead = 1 - m_aucRead;
    }
    ++m_uiPopCount;
    ++m_uiReadPos;
  }

  T& at( size_t pos )
  {
    if( pos < m_uiPopCount )
    {
      throw AccessError();
    }
    pos -= m_uiPopCount;
    if( pos >= size() )
    {
      throw AccessError();
    }
    const size_t uiRemainderInRead = m_aacT[m_aucRead].size() - m_uiReadPos;
    if( pos >= uiRemainderInRead )
    {
      return m_aacT[1-m_aucRead][pos - uiRemainderInRead];
    }
    else
    {
      return m_aacT[m_aucRead][pos + m_uiReadPos];
    }
  }

  T& front() { return at( m_uiPopCount ); }
//  T& back()  { return at( m_uiPopCount + size() - 1 ); }

private:
  std::vector<T>* m_aacT;
  UChar           m_aucWrite;
  UChar           m_aucRead;
  size_t          m_uiReadPos;
  size_t          m_uiWritePos;
  size_t          m_uiPopCount;
};

#endif

