/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TEncSbac.h
    \brief    Context-adaptive entropy encoder class (header)
*/

#ifndef __TENCSBAC__
#define __TENCSBAC__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "../TLibCommon/TComBitStream.h"
#include "../TLibCommon/ContextTables.h"
#include "../TLibCommon/ContextModel.h"
#include "../TLibCommon/ContextModel3DBuffer.h"
#include "TEncEntropy.h"
#include "TEncBinCoder.h"
#include "TEncBinCoderCABAC.h"

class TEncTop;

#if HHI_SYS_PIPE
#include <vector>
#include <map>
#include "../TLibCommon/Queue.h"

#if 0
#define COUNT_BINS if( m_bBinCountingEnabled ) m_uiBinsCoded++;
#else
#define COUNT_BINS
#endif

class CabacFinishError {};

class CodewordInterleaver
{
public:
  CodewordInterleaver() : m_pcBitIf( NULL ) {}

  Void setBitIf( TComBitIf *p ) { m_pcBitIf = p; }

  Void resetInterleaver()
  {
    m_auiCodewords.reset();
    m_aucLengths.reset();
  }

  UInt reserveCodeword()
  {
    m_aucLengths.push( 255 );
    return UInt( m_auiCodewords.push( 0 ) );
  }

  Void writeCodeword( UInt uiId, UInt uiBits, UChar ucLength )
  {
    DTRACE_CABAC_T( "CW=" )
    DTRACE_CABAC_V( uiId )
    DTRACE_CABAC_T( "\tbits=" )
    DTRACE_CABAC_X( uiBits )
    DTRACE_CABAC_T( "\tlen=" )
    DTRACE_CABAC_V( ucLength )
    DTRACE_CABAC_T( "\n" )
    assert( ucLength <= 32 );
    m_auiCodewords.at( uiId ) = uiBits;
    m_aucLengths.at( uiId ) = ucLength;
    xCheckAndWrite();
  }

private:
  Void xCheckAndWrite()
  {
    while( m_aucLengths.size() != 0 && m_aucLengths.front() != 255 )
    {
      m_pcBitIf->write( m_auiCodewords.front(), m_aucLengths.front() );
      m_aucLengths.pop();
      m_auiCodewords.pop();
    }
  }

  TComBitIf*   m_pcBitIf;
  Queue<UInt>  m_auiCodewords;
  Queue<UChar> m_aucLengths;
};

class V2VEncoder
{
public:
  V2VEncoder() : m_pcInterleaver( NULL ) {}
  Void setInterleaver( CodewordInterleaver& rcInterleaver ) { m_pcInterleaver = &rcInterleaver; }

  Void init( const Char * paacV2VCode[][2], size_t uiSizeofCode )
  {
    if( !m_auiV2VCode.empty() )
    {
      throw;
    }
    const UInt uiNumCodewords = UInt( uiSizeofCode / sizeof( const Char* ) / 2 );
    for( UInt ui = 0; ui < uiNumCodewords; ++ui )
    {
      UInt64 uiPrim = xGetCodeFromString( paacV2VCode[ui][0] );
      UInt64 uiSeco = xGetCodeFromString( paacV2VCode[ui][1] );
      if( m_auiV2VCode.find( uiPrim ) != m_auiV2VCode.end() )
      {
        throw;
      }
      m_auiV2VCode.insert( std::make_pair( uiPrim, uiSeco ) );
    }
  }
  UInt  m_uiCw;
  Void (V2VEncoder::*encode1BinTab[8])( UInt uiBin, ContextModel &rcCtxModel );
  Void (V2VEncoder::*finishCodewordTab[8])();
  Void  xFinishCodeword0();
  Void  xFinishCodeword1();
  Void  xFinishCodeword2();
  Void  xFinishCodeword3();
  Void  xFinishCodeword4();
  Void  xFinishCodeword5();
  Void  xFinishCodeword6();
  Void  xFinishCodeword7();
  Void  xEncode1Bin0       ( UInt  uiBin,  ContextModel& rcCtxModel );
  Void  xEncode1Bin1       ( UInt  uiBin,  ContextModel& rcCtxModel );
  Void  xEncode1Bin2       ( UInt  uiBin,  ContextModel& rcCtxModel );
  Void  xEncode1Bin3       ( UInt  uiBin,  ContextModel& rcCtxModel );
  Void  xEncode1Bin4       ( UInt  uiBin,  ContextModel& rcCtxModel );
  Void  xEncode1Bin5       ( UInt  uiBin,  ContextModel& rcCtxModel );
  Void  xEncode1Bin6       ( UInt  uiBin,  ContextModel& rcCtxModel );
  Void  xEncode1Bin7       ( UInt  uiBin,  ContextModel& rcCtxModel );
  Void  encodeBinUR5       ( UInt uiBin );

  Void (V2VEncoder::*encode1BinPtr)( UInt uiBin, ContextModel &rcCtxModel );
  Void (V2VEncoder::*finishCodewordPtr)();

  Void init( Void         (V2VEncoder::*encode1BinFunc)( UInt, ContextModel& ),
             Void         (V2VEncoder::*finishCodewordFunc)(),
             UInt         uiPipeId )
  {
    m_uiPipeId        = uiPipeId;
    encode1BinPtr     = encode1BinFunc;
    finishCodewordPtr = finishCodewordFunc;
    resetCoder();
  }

  __inline Void encode1Bin( UInt uiBin, ContextModel &rcCtxModel )
  {
    (this->*encode1BinPtr)( uiBin, rcCtxModel );
  }
  __inline Void finishCodeword()
  {
    (this->*finishCodewordPtr)();
  }
  __inline Void encodeBinEP( UInt uiBin )
  {
    m_auiCwIdx = m_pcInterleaver->reserveCodeword();
    m_pcInterleaver->writeCodeword( m_auiCwIdx, uiBin, 1 ); 
  }


  Void resetCoder()
  {
    m_uiLengthCurr = 0;
    m_uiCodewordCurr = 0;
    m_uiCw = ( m_uiPipeId == PIPE_ID_TB || m_uiPipeId == PIPE_ID_BP3 ) ? 1 : 0;
  }

  Void encodeBin( UInt uiBin )
  {
    {
      //DTRACE_CABAC_V( g_nSymbolCounter++ )
      //DTRACE_CABAC_T( "\tPIPE " )
    }
    if( m_uiLengthCurr == 0 )
    {
      {
        //DTRACE_CABAC_T( "\tread cw " )
      }
      m_auiCwIdx = m_pcInterleaver->reserveCodeword();
    }
    {
      //DTRACE_CABAC_T( "bin = " )
      //DTRACE_CABAC_V( uiBin )
      //DTRACE_CABAC_T( "\n" )
    }
    ++m_uiLengthCurr;
    m_uiCodewordCurr += m_uiCodewordCurr + (uiBin ? 1: 0);
    UInt64 uiCw = ( m_uiLengthCurr << m_ucLengthShift ) | m_uiCodewordCurr;
    std::map<UInt64, UInt64>::iterator iter = m_auiV2VCode.find( uiCw );
    if( iter != m_auiV2VCode.end() )
    {
      //DTRACE_CABAC_T( "cw(" );
      //DTRACE_CABAC_V( UInt( iter->second & m_uiCWMask ) );
      //DTRACE_CABAC_T( "," );
      //DTRACE_CABAC_V( UChar( iter->second >> m_ucLengthShift ) );
      //DTRACE_CABAC_T( ")" );
      m_pcInterleaver->writeCodeword( m_auiCwIdx, UInt( iter->second & m_uiCWMask ), UChar( iter->second >> m_ucLengthShift ) );
      m_uiLengthCurr = 0;
      m_uiCodewordCurr = 0;
    }
  }
private:
  UInt64 xGetCodeFromString( const Char *p )
  {
    UInt64 uiLength = 0;
    UInt64 uiValue = 0;
    while( p[uiLength] != '\0' )
    {
      uiValue += uiValue;
      if( p[uiLength] == '1' )
      {
        ++uiValue;
      }
      else if( p[uiLength] != '0' )
      {
        throw;
      }
      uiLength++;
    }
    if( uiLength > m_ucLengthShift )
      throw;
    return ( uiLength << m_ucLengthShift ) | uiValue;
  }

  static const UChar       m_ucLengthShift = 58;
  static const UInt64      m_uiCWMask = ( UInt64( 1 ) << 58 ) - 1;
  std::map<UInt64, UInt64> m_auiV2VCode;
  UInt64                   m_uiLengthCurr;
  UInt64                   m_uiCodewordCurr;
  CodewordInterleaver*     m_pcInterleaver;
  UInt                     m_auiCwIdx;
  UInt                     m_uiPipeId;
};
#endif

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// SBAC encoder class
class TEncSbac : public TEncEntropyIf
{
public:
#if HHI_SYS_PIPE
  TEncSbac( Bool bRD = false );
#else
  TEncSbac();
#endif
  virtual ~TEncSbac();

#if !HHI_SYS_PIPE
  Void  init                   ( TEncBinIf* p )  { m_pcBinIf = p; }
  Void  uninit                 ()                { m_pcBinIf = 0; }
#endif
  
  //  Virtual list
  Void  resetEntropy           ();
#if HHI_SYS_PIPE
  Void  setBitstream           ( TComBitIf* p )  { m_pcBitIf = p;
    if( !m_bEst )
    {
      m_cInterleaver.setBitIf( m_pcBitIf );
    }
  }
#else
  Void  setBitstream           ( TComBitIf* p )  { m_pcBitIf = p; m_pcBinIf->init( p ); }
#endif
  Void  setSlice               ( TComSlice* p )  { m_pcSlice = p;                       }
  
  Bool  getAlfCtrl             ()                         { return m_bAlfCtrl;          }
  UInt  getMaxAlfCtrlDepth     ()                         { return m_uiMaxAlfCtrlDepth; }
  Void  setAlfCtrl             ( Bool bAlfCtrl          ) { m_bAlfCtrl          = bAlfCtrl;          }
  Void  setMaxAlfCtrlDepth     ( UInt uiMaxAlfCtrlDepth ) { m_uiMaxAlfCtrlDepth = uiMaxAlfCtrlDepth; }
  
  // SBAC RD
  Void  resetCoeffCost         ()                { m_uiCoeffCost = 0;  }
  UInt  getCoeffCost           ()                { return  m_uiCoeffCost;  }
  
  Void  load                   ( TEncSbac* pScr  );
  Void  loadIntraDirModeLuma   ( TEncSbac* pScr  );
  Void  store                  ( TEncSbac* pDest );
#if HHI_SYS_PIPE
  Void  resetBits              ()                { xResetBits(); }
  UInt  getNumberOfWrittenBits ()                { return xGetNumWrittenBits(); }
#else
  Void  resetBits              ()                { m_pcBinIf->resetBits(); m_pcBitIf->resetBits(); }
  UInt  getNumberOfWrittenBits ()                { return m_pcBinIf->getNumWrittenBits(); }
#endif
  //--SBAC RD

  Void  codeSPS                 ( TComSPS* pcSPS     );
  Void  codePPS                 ( TComPPS* pcPPS     );
  void  codeSEI                 ( const SEI&         );
  Void  codeSliceHeader         ( TComSlice* pcSlice );
  Void  codeTerminatingBit      ( UInt uilsLast      );
  Void  codeSliceFinish         ();
  
  Void  codeAlfFlag       ( UInt uiCode );
  Void  codeAlfUvlc       ( UInt uiCode );
  Void  codeAlfSvlc       ( Int  uiCode );
  Void  codeAlfCtrlDepth  ();
#if TSB_ALF_HEADER

#if MTK_NONCROSS_INLOOP_FILTER
  /// Code number of ALF CU control flags
  Void codeAlfFlagNum        ( UInt uiCode, UInt minValue, Int iDepth);
#else
  Void codeAlfFlagNum        ( UInt uiCode, UInt minValue );
#endif

  Void codeAlfCtrlFlag       ( UInt uiSymbol );
#endif
#if MTK_SAO
  Void  codeAoFlag       ( UInt uiCode );
  Void  codeAoUvlc       ( UInt uiCode );
  Void  codeAoSvlc       ( Int  uiCode );
#endif

private:
  Void  xWriteUnarySymbol    ( UInt uiSymbol, ContextModel* pcSCModel, Int iOffset );
  Void  xWriteUnaryMaxSymbol ( UInt uiSymbol, ContextModel* pcSCModel, Int iOffset, UInt uiMaxSymbol );
  Void  xWriteEpExGolomb     ( UInt uiSymbol, UInt uiCount );
#if E253
  Void  xWriteGoRiceExGolomb ( UInt uiSymbol, UInt &ruiGoRiceParam );
#else
  Void  xWriteExGolombLevel  ( UInt uiSymbol, ContextModel& rcSCModel  );
#endif
  Void  xWriteTerminatingBit ( UInt uiBit );
  
#if MVD_CTX
  Void  xWriteMvd            ( Int iMvd, UInt uiAbsSumL, UInt uiAbsSumA, UInt uiCtx );
#else
  Void  xWriteMvd            ( Int iMvd, UInt uiAbsSum, UInt uiCtx );
#endif
  Void  xWriteExGolombMvd    ( UInt uiSymbol, ContextModel* pcSCModel, UInt uiMaxBin );
  Void  xCopyFrom            ( TEncSbac* pSrc );
  
protected:
  TComSlice*    m_pcSlice;
  TComBitIf*    m_pcBitIf;
#if HHI_SYS_PIPE
  const Bool    m_bEst;
  Bool          m_bBinCountingEnabled;
#else
  TEncBinIf*    m_pcBinIf;
#endif
  Bool          m_bAlfCtrl;
  
  //SBAC RD
  UInt          m_uiCoeffCost;
  
#if HHI_SYS_PIPE_LC_MODE
  Int           m_initQp;
  SliceType     m_initSliceType;
#endif
  
  // Adaptive loop filter
  UInt          m_uiMaxAlfCtrlDepth;
#if FINE_GRANULARITY_SLICES && MTK_NONCROSS_INLOOP_FILTER
  Int           m_iSliceGranularity; //!< slice granularity
#endif
  //--Adaptive loop filter
  
public:

#if FINE_GRANULARITY_SLICES && MTK_NONCROSS_INLOOP_FILTER
  /// set slice granularity
  Void setSliceGranularity(Int iSliceGranularity)  {m_iSliceGranularity = iSliceGranularity;}

  /// get slice granularity
  Int  getSliceGranularity()                       {return m_iSliceGranularity;             }
#endif
  Void codeAlfCtrlFlag   ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  Void codeSkipFlag      ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  Void codeMergeFlag     ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  Void codeMergeIndex    ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  Void codeSplitFlag     ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  Void codeMVPIdx        ( TComDataCU* pcCU, UInt uiAbsPartIdx, RefPicList eRefList );
  
  Void codePartSize      ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth );
  Void codePredMode      ( TComDataCU* pcCU, UInt uiAbsPartIdx );
#if E057_INTRA_PCM
  Void codeIPCMInfo      ( TComDataCU* pcCU, UInt uiAbsPartIdx );
#endif
  Void codeTransformSubdivFlag ( UInt uiSymbol, UInt uiCtx );
  Void codeQtCbf               ( TComDataCU* pcCU, UInt uiAbsPartIdx, TextType eType, UInt uiTrDepth );
  Void codeQtRootCbf           ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  
  Void codeIntraDirLumaAng     ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  
  Void codeIntraDirChroma      ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  Void codeInterDir            ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  Void codeRefFrmIdx           ( TComDataCU* pcCU, UInt uiAbsPartIdx, RefPicList eRefList );
  Void codeMvd                 ( TComDataCU* pcCU, UInt uiAbsPartIdx, RefPicList eRefList );
  
  Void codeDeltaQP             ( TComDataCU* pcCU, UInt uiAbsPartIdx );
  Void codeCbf                 ( TComDataCU* pcCU, UInt uiAbsPartIdx, TextType eType, UInt uiTrDepth ) {}
  Void codeBlockCbf            ( TComDataCU* pcCU, UInt uiAbsPartIdx, TextType eType, UInt uiTrDepth, UInt uiQPartNum, Bool bRD = false ) {}
  
#if CAVLC_RQT_CBP
  Void codeCbfTrdiv      ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth ) {}
  UInt xGetFlagPattern   ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth ) { return 0; }
#endif
#if PCP_SIGMAP_SIMPLE_LAST
  __inline Void codeLastSignificantXY ( UInt uiPosX, UInt uiPosY, const UInt uiWidth, const TextType eTType, const UInt uiCTXIdx, const UInt uiScanIdx );
#endif
#if HHI_SYS_PIPE
#if HHI_SYS_PIPE_LC_MODE // TODO here
  Void  xWriteLastPos     ( UInt uiPosX, UInt uiPosY, const UInt uiWidth, const TextType eTType, const UInt uiCTXIdx, const UInt uiScanIdx );
#else
  Void  xCodeRemainder( UInt uiScanPosLast, UInt uiScanIdx, UInt uiLog2BlockSize, TCoeff *pcCoef, TextType eTType );
#endif
#endif
  Void codeCoeffNxN            ( TComDataCU* pcCU, TCoeff* pcCoef, UInt uiAbsPartIdx, UInt uiWidth, UInt uiHeight, UInt uiDepth, TextType eTType, Bool bRD = false );
  
  // -------------------------------------------------------------------------------------------------------------------
  // for RD-optimizatioon
  // -------------------------------------------------------------------------------------------------------------------
  
  Void estBit                        ( estBitsSbacStruct* pcEstBitsSbac, UInt uiCTXIdx, TextType eTType );
  Void estCBFBit                     ( estBitsSbacStruct* pcEstBitsSbac, UInt uiCTXIdx, TextType eTType );
  Void estSignificantMapBit          ( estBitsSbacStruct* pcEstBitsSbac, UInt uiCTXIdx, TextType eTType );
  Void estSignificantCoefficientsBit ( estBitsSbacStruct* pcEstBitsSbac, UInt uiCTXIdx, TextType eTType );
  
  __inline Int  biari_no_bits        ( Short symbol, ContextModel& rcSCModel );
  
#if !HHI_SYS_PIPE
  TEncBinIf* getEncBinIf()  { return m_pcBinIf; }
#endif
private:
#if HHI_SYS_PIPE
  UInt64               m_uiBitsCounter;
#endif
  UInt                 m_uiLastQp;
  ContextModel3DBuffer m_cCUSplitFlagSCModel;
  ContextModel3DBuffer m_cCUSkipFlagSCModel;
  ContextModel3DBuffer m_cCUMergeFlagExtSCModel;
  ContextModel3DBuffer m_cCUMergeIdxExtSCModel;
  ContextModel3DBuffer m_cCUPartSizeSCModel;
  ContextModel3DBuffer m_cCUPredModeSCModel;
  ContextModel3DBuffer m_cCUAlfCtrlFlagSCModel;
  ContextModel3DBuffer m_cCUIntraPredSCModel;
#if ADD_PLANAR_MODE
  ContextModel3DBuffer m_cPlanarFlagSCModel;
#endif
  ContextModel3DBuffer m_cCUChromaPredSCModel;
  ContextModel3DBuffer m_cCUDeltaQpSCModel;
  ContextModel3DBuffer m_cCUInterDirSCModel;
  ContextModel3DBuffer m_cCURefPicSCModel;
  ContextModel3DBuffer m_cCUMvdSCModel;
  ContextModel3DBuffer m_cCUQtCbfSCModel;
  ContextModel3DBuffer m_cCUTransSubdivFlagSCModel;
  ContextModel3DBuffer m_cCUQtRootCbfSCModel;
  
  ContextModel3DBuffer m_cCUSigSCModel;
#if PCP_SIGMAP_SIMPLE_LAST
  ContextModel3DBuffer m_cCuCtxLastX;
  ContextModel3DBuffer m_cCuCtxLastY;
#else
  ContextModel3DBuffer m_cCULastSCModel;
#endif
  ContextModel3DBuffer m_cCUOneSCModel;
  ContextModel3DBuffer m_cCUAbsSCModel;
  
  ContextModel3DBuffer m_cMVPIdxSCModel;
  
  ContextModel3DBuffer m_cALFFlagSCModel;
  ContextModel3DBuffer m_cALFUvlcSCModel;
  ContextModel3DBuffer m_cALFSvlcSCModel;
#if MTK_SAO
  ContextModel3DBuffer m_cAOFlagSCModel;
  ContextModel3DBuffer m_cAOUvlcSCModel;
  ContextModel3DBuffer m_cAOSvlcSCModel;
#endif

  // need to move
#if HHI_SYS_PIPE
  CodewordInterleaver m_cInterleaver;
  V2VEncoder          m_acV2VEncoder[8];
  UInt                m_uiBinsCoded;

  Void  xStart             ();
  Void  xFinish            ();
  Void  xCopyState         ( TEncSbac *pSrc );

  Void  xResetBits         ();
  Void  xResetBitstreamIf  ();
  UInt  xGetNumWrittenBits ();

  Void  xEncodeBin         ( UInt  uiBin,  ContextModel& rcCtxModel );
  __inline Void xEncodeBinEP( UInt  uiBin                            )
  {
    COUNT_BINS
    if( !m_bEst )
    {
      m_acV2VEncoder[PIPE_ID_EP].encodeBinEP( uiBin );
      return;
    }
    m_uiBitsCounter += 1 << 15;
  }
  Void  xEncodeBinTrm      ( UInt  uiBin                            );

public:
  Void  setBinsCoded              ( UInt uiVal )  { m_uiBinsCoded = uiVal;          }
  UInt  getBinsCoded              ()              { return m_uiBinsCoded;           }
  Void  setBinCountingEnableFlag  ( Bool bFlag )  { m_bBinCountingEnabled = bFlag;  }
  Bool  getBinCountingEnableFlag  ()              { return m_bBinCountingEnabled;   }

protected:
#endif
};

#endif // !defined(AFX_TENCSBAC_H__DDA7CDC4_EDE3_4015_9D32_2156249C82AA__INCLUDED_)
